__version__ = '0.5.5'
__author__ = 'John Connell - The Mathematic Fanatic'
__description__ = 'Manim plugin which aims to make it much easier to meaningfully transform algebra expressions.'

from .expressions import *
from .actions import *
from .timelines import *
from .utils import *
from .algebra import *
from .calculus import *
from .trigonometry import *
from .sets import *
